/********************************************************************
Implementation of algorithm UnitWalk 
by Edward A. Hirsch and Arist Kojevnikov, 2001.
This file contains procedures which are not specific 
for a concrete algorithm.
*********************************************************************/

#include "framework.h"

#define DEF_I (1000*1000*1000)
//max number of algorithm runs
#define DEF_T (30*7*24*60*60)
//max time of algorithm work
#define DEF_A (1000*1000*1000)
//max number of assignments
#define DEF_p -1
//max number of flips, +\infty can be defined by -1 

////////////////////////////////////////////////////////////////////////
char * itoa(number x)
{
	char*s=(char*)malloc(32);
	snprintf(s,31,"%lld",x);
	return s;
}

////////////////////////////////////////////////////////////////////////
number dist(number *Vector,number *Vector1) 
//returns Hamming distance between Vector and Vector1
{
  number Counter, Diff = 0;

#ifndef ___MYVAR_DEBUG___     
  for(Counter=0;Counter<NumberOfVars;Counter++)
  {
    if(Vector[Counter]!=Vector1[Counter])
      Diff++;  
  }
#endif

#ifdef ___MYVAR_DEBUG___       
//some debug information 
//that output on screen  
  for(Counter=0;Counter<NumberOfVars;Counter++)
  {
    if(Vector[Counter]!=Vector1[Counter])
    {
      Diff++; 
      printf("|%lld|",Vector[Counter]); 
    }
    else
      printf("%lld",Vector[Counter]); 
    
  }
  
  for(Counter=0;Counter<NumberOfVars;Counter++)
  {
    if(Vector[Counter]!=Vector1[Counter])
      printf("|%lld|",Vector1[Counter]); 
    else
      printf("%lld",Vector1[Counter]); 
   
  }
  
#endif
 return Diff;
 
}
////////////////////////////////////////////////////////////////////////
void OutVariablesInClauses(void) 
// prints the current formula 
{
  number Counter;
  number VarC;
  struct MyList *ThisClause;
  
  for(VarC=0;VarC<NumberOfVars;VarC++)
  {
    printf("\n%lld: ", VarC);
    ThisClause = VariablesInClauses[VarC];
 
    while(ThisClause!=NULL)
    {
      printf("%lld ",ThisClause->Number);
      ThisClause = ThisClause->link;
    }
  }
}

////////////////////////////////////////////////////////////////////////
void FreeVariablesInClauses()
 // free used memory after the execution
{
  number Counter;
  struct MyList *ThisClause;
  struct MyList *NextClause;

  ThisClause = PreDualClauses;

  while(ThisClause!=NULL){
    NextClause = ThisClause->link;
    free(ThisClause);
    ThisClause = NextClause;
  }

  
  for(Counter=0;Counter<NumberOfVars;Counter++)
  {
    ThisClause = VariablesInClauses[Counter];
  
    while(ThisClause!=NULL)
    {
      NextClause = ThisClause->link;
      free(ThisClause);
      ThisClause = NextClause;

    }
  }
}

/* ///////////////////////Satisfy /////////////////////////////////
This function check formula for satisfaction with vector of variables*/
number Satisfy(struct NumberWithLink *Formula, number *Vector)
{
  number Counter1, Counter2, VarNumber;
  number ValueOfClause = 0;
  number NumberOfVarsInClause;
  number *Clause;
  
  for(Counter1=0;Counter1<NumberOfClauses;Counter1++)
  {
    NumberOfVarsInClause = SavedSOC[Counter1];
    Clause = Formula[Counter1].link;
    
    for(Counter2=0;Counter2<NumberOfVarsInClause;Counter2++)
    {
      VarNumber=Clause[Counter2];
      if(VarNumber<0)
      {
        if(Vector[-VarNumber - 1]==0)
      	{
      	  ValueOfClause = 1;
      	  break;
      	}
      }
      else if(VarNumber>0)
      {
      	if(Vector[VarNumber - 1]==1)
      	{
      	  ValueOfClause = 1;
      	  break;
      	}    
      }
    }
    
    if(ValueOfClause==0)
    {
      return 0;
    }
    
    ValueOfClause=0;
  }
  
  return 1;
}

////////////////////////////////////////////////////////////////////////
void FormulaOut(struct NumberWithLink *Formula) 
// prints the current formula in dimacs format
{
  number Counter,Counter1, NumberOfVarsInClause;
  number *Clause;
  
  printf("\np cnf %lld %lld\n",NumberOfVars,NumberOfClauses);
  
  for(Counter=0;Counter<NumberOfClauses;Counter++)
  {
    Clause = Formula[Counter].link;
    NumberOfVarsInClause = SavedSOC[Counter];
    
    for(Counter1=0;Counter1<NumberOfVarsInClause;Counter1++)
    {
      printf("%lld ",Clause[Counter1]);      
    }
    if(NumberOfVarsInClause!=0)
      printf("0\n");      
  }
}

/*
 * Functions for work with random strings :
 */

////////////////////////////////////////////////////////////////////////
void SetPermutation(number *Permutation) 
// fill a trivial permutation
{
  number Counter;

  for(Counter=0;Counter<NumberOfVars;Counter++)
  {
    Permutation[Counter]=Counter+1;
  }
}


////////////////////////////////////////////////////////////////////////
void GetRandomPermutation(number *Permutation) 
// fill a random permutation
{
  number Random,Counter,Number;

  for(Counter=0;Counter<NumberOfVars;Counter++)
  {
    Random = random()%NumberOfVars;
    
    Number=Permutation[Counter];
    Permutation[Counter]=Permutation[Random];
    Permutation[Random]=Number;
  }
}


////////////////////////////////////////////////////////////////////////
void GetRandomVector(number *Vector) 
// fill a random vector from {0,1}^n
{
  number Random;
  number Counter;

  for(Counter=0;Counter<NumberOfVars;Counter++)
  {
    Random = random()%2;
    Vector[Counter]=Random;
  }
}

////////////////////////////////////////////////////////////////////////
void PutNumbers(struct NumberWithLink *CNFF)
//reads VariablesInClauses and CNFF from file
{
  number Counter1,Counter2,Counter3,ScanedNumber, Clause[NumberOfVars];
  number SizeOfNumber = sizeof(number);
  struct MyList *ThisClause;
  number sizeOfList = sizeof(struct MyList);
  struct MyList *TwoClause; 

  for(Counter1=0;Counter1<NumberOfClauses;Counter1++)
  {
    for(Counter2=0;Counter2<NumberOfVars+1;Counter2++)
    {
      fscanf(inFile, "%lld", &ScanedNumber);
      
      if(ScanedNumber!=0)
      {
// beforehand scan information of one clause from input file
        Clause[Counter2] = ScanedNumber;   
        if(ScanedNumber>0)
        {
        //Structure VariablesInClauses is filled
        /* 
          VariablesInClauses is Array of lists. VariablesInClauses[k-1] 
          is list of clauses numbers with variable x_k. This array is 
          need for quick search for clause with some variable.
        */ 
          ThisClause = (struct MyList *) malloc(sizeOfList);
          ThisClause->link   = VariablesInClauses[ScanedNumber-1];
          ThisClause->Number = Counter1;          
          VariablesInClauses[ScanedNumber-1]=ThisClause;
        }
        else
        {
          //Structure VariablesInClauses is filled
          ThisClause = (struct MyList *) malloc(sizeOfList);
          ThisClause->link = VariablesInClauses[-ScanedNumber-1];
          ThisClause->Number = Counter1;       
          VariablesInClauses[-ScanedNumber-1]=ThisClause;
        }
      }
      else 
      //The end of clause's reading from input  
      {
        // if clause is unit then add it to PreUnitClauses
        if(Counter2 == 1)
        {
          PreUnitClauses[PreNumberOfUnitClauses]=Clause[0];
          PreNumberOfUnitClauses++;
        }
        // if clause have size 2 then add is to PreDualClauses list
        else if(Counter2 == 2){
          TwoClause         = (struct MyList *) malloc(sizeOfList);
          TwoClause->link   = PreDualClauses;
          TwoClause->Number = Counter1;
          PreDualClauses  = TwoClause;
        }

        // fill data of clause
        SavedSOC[Counter1] = Counter2; 
        CNFF[Counter1].link = (number *) malloc(SizeOfNumber*Counter2);    
        for(Counter3=0;Counter3<Counter2;Counter3++)
          CNFF[Counter1].link[Counter3] = Clause[Counter3]; 
        break;
      }
    }
    while(ScanedNumber!=0)
      fscanf(inFile,"%lld", &ScanedNumber);
  }
}

////////////////////////////////////////////////////////////////////////
number WorkWithCNF() 
// process the body of .cnf-file, initialize our data and call Search
{
  number ValueToReturn;
  struct NumberWithLink CNFF[NumberOfClauses]; 
  struct MyList *z[NumberOfVars];
  number zh[2*NumberOfVars], temp[NumberOfVars], unit[NumberOfVars];
  number soc[NumberOfClauses];
  
  //inition of all global variables
  VariablesInClauses = z;
#ifdef _BIN_
  tempAssign = temp;
#endif
  United = unit;
  
  for(ValueToReturn=0;ValueToReturn<NumberOfVars;ValueToReturn++){
    VariablesInClauses[ValueToReturn]=NULL;
  }
  SavedSOC = soc;
  
  PreDualClauses = NULL;  
  PreNumberOfUnitClauses = 0;
  PreUnitClauses = zh;
  
  PutNumbers(CNFF);
  
  fclose(inFile);

//  FormulaOut(CNFF);      
  ValueToReturn=Search(CNFF);
  return ValueToReturn;
}


////////////////////////////////////////////////////////////////////////
void WorkWithSatAssignment(char *InputFile)
// process the body of file with satisfying assignment, 
// initialize our data and call Search
{
  FILE *mainFile = fopen(InputFile, "r");
  char Symbol, NameOfCNFFile[256], v[1];
  number i;

  if(mainFile==NULL)
  {
    printf("c Cannot open input file.\n");
    exit(-1);
  }

  fscanf(mainFile, "%c", &Symbol);

  while(1)
  {
  
    while((Symbol!='f')&&(Symbol!=EOF))
    {
      while((Symbol!='\n')&&(Symbol!=EOF))
        fscanf(mainFile,"%c", &Symbol);
      fscanf(mainFile,"%c", &Symbol);
    }
  
    if (Symbol==EOF) break;
    
    fscanf(mainFile,"%s", NameOfCNFFile);

    if (!Quiet) printf("f %s",NameOfCNFFile);   
    fprintf(outFile,"f %s",NameOfCNFFile);  
    fflush(outFile); 

    WorkWithFile(NameOfCNFFile);
   

    SatAssignment = (number *)malloc(sizeof(number)*NumberOfVars);
    
    fscanf(mainFile,"%c", &Symbol);
    while(Symbol=='v'||Symbol==' ')
      fscanf(mainFile,"%c", &Symbol);
  
    i=0;  
    while(Symbol=='0'||Symbol=='1')  
    {
      SatAssignment[i]=Symbol-'0';

      fscanf(mainFile,"%c", &Symbol);
      i++;
    }

    WorkWithCNF();
    
    free(SatAssignment);

    while((Symbol!='\n')&&(Symbol!=EOF))
	  fscanf(mainFile,"%c", &Symbol);
	
	fscanf(mainFile,"%c", &Symbol);
	
	while((Symbol!='\n')&&(Symbol!=EOF))
      fscanf(mainFile,"%c", &Symbol);

    fscanf(mainFile,"%c", &Symbol);
    
    
  }  
  fclose(mainFile);
}


////////////////////////////////////////////////////////////////////////
void WorkWithFile(char *InputFile) 
// process the head of .cnf-file
// and call WorkWithCNF for further
// processing
{
  number Counter1, Counter2, ScanedNumber;
  char Symbol, *cnf[3];
  
  inFile = fopen(InputFile, "r");
  if(!inFile)
  {
    printf("c Cannot open input file : %s\n", InputFile);
    exit(-1);
  }
  
  // eat everything before first ^p
  fscanf(inFile, "%c", &Symbol);
  while(Symbol!='p')
  {
    while(Symbol!='\n')
      fscanf(inFile,"%c", &Symbol);
    fscanf(inFile,"%c", &Symbol);
  }
  
  // read the string "cnf" and number of variables and number of clauses
  fscanf(inFile,"%s%lld%lld", cnf, &NumberOfVars, &NumberOfClauses);
#ifdef _SAT_COMP_
  printf("c Read formula with %lld variables and %lld clauses\n",NumberOfVars,
  NumberOfClauses);
#endif
}

////////////////////////////////////////////////////////////////////////
int listdir(char *dirname) 
// work with directory of .cnf-files
{
   int i;
   char *NameOfFile, Character, FullNameOffile[512];

   register struct dirent *dirbuf;
   DIR *fddir;
   ino_t dot_ino=0, dotdot_ino=0;

  if((fddir=opendir(dirname))==NULL)
  {
    fprintf(stderr,"Can't read %s\n",dirname);
    return 1;
  }
  
  /* without alphabetical sorting  */
  while ((dirbuf=readdir(fddir))!=NULL ) 
  {
    if (dirbuf->d_ino==0) 
      continue;
    if (strcmp(dirbuf->d_name,".")==0)
    {
      dot_ino=dirbuf->d_ino;
      continue;
    } 
    else if(strcmp(dirbuf->d_name,"..")==0)
    {
      dotdot_ino = dirbuf -> d_ino ;
     continue;
   }
   else 
     NameOfFile =  dirbuf -> d_name;
     for(i=0;(Character=NameOfFile[i])!='\0';i++)
     {
/*
       if(Character=='.')
       if(NameOfFile[i+1]=='c')
       if(NameOfFile[i+2]=='n')
       if(NameOfFile[i+3]=='f')
*/     
       {  
         if (!Quiet) printf("f ");
         fprintf(outFile,"f ");
         *FullNameOffile='\0';
         strcat(FullNameOffile,dirname);
         strcat(FullNameOffile,"/");
         strcat(FullNameOffile,NameOfFile);
         if (!Quiet) printf("%s ", FullNameOffile);
	       fflush(stdout);
         fprintf(outFile,"%s ", FullNameOffile);
	       fflush(outFile);
         WorkWithFile(FullNameOffile);
         WorkWithCNF(inFile);
         fflush(outFile);
       }     
     }
  }
  
  closedir(fddir);

  if(dot_ino==0) 
    printf("Bad directory: no \".\"\n");
  if(dotdot_ino==0) 
    printf("Bad directory: no \"..\"\n");
  if((dot_ino)&&(dot_ino==dotdot_ino)) 
    printf("This is the root directory.\n");

  return 0;
}


////////////////////////////////////////////////////////////////////////
void help() 
// Give help and exit
{
    printf("\n");
    printf("General parameters:\n");
    printf("\n");
    printf("  -f file = name of input file\n");
    printf("  -d dir  = name of input directory\n");
    printf("  -s seed = seed\n");
    printf("  -sr     = random seed\n");
    printf("  -o file = name of output file\n");
    printf("  -a file = name of file with satisfying assignments\n");
    printf("  -I I    = runs limit\n");
    printf("  -T T    = time limit in seconds\n");
    printf("  -A A    = assigns limit\n");
    printf("  -p p    = flips limit\n");
    printf("  -q      = be quiet\n");
    printf("  -h      = this help\n");
    printf("  -r      = number of restarts\n");
    printf("\n");
    printf("For example:\n");
    printf("\n");
    printf("  ./UnitWalk -d cnf/uf800 -T 60\n");
    printf("\n");

    printf("The default values are -I %lld, -T %lld, -A %lld, -p %lld, -r 1, -s 28122101.\n",DEF_I,DEF_T,DEF_A,DEF_p);   
    exit(-1); 
}

/* 
 * Main Program :
 */
#ifdef _SAT_COMP_
// for a SAT2002 competition
int main(int argc,char *argv[]){
  double result;
//  double result;
  char *NameOfFile;
  int CheckInDir=0;
  // will check all .cnf-files in a directory?
        
  if(argc != 3){     
    printf("c Wrong command line format!\n");
    exit(-3);  
  }

  NameOfFile = argv[1];

  result = atof(argv[2]);
  
  printf("c UnitWalk starting with random seed %.0f on %s\n",result ,NameOfFile);
    
  srandom(result);
//  printf("c UnitWalk starting with random seed %lld on %s\nc Trying to find a solution...\n",result ,NameOfFile);
      
  //default values
  I = DEF_I;
  T = DEF_T;
  A = DEF_A;
  p = DEF_p;
  Quiet = 1;
  aGiven = 0;
  MaxTry = 1;
  
//  MedianFlip = (number*)malloc(MaxTry*sizeof(number));
//  MedianAssign = (number*)malloc(MaxTry*sizeof(number));
 
 
  WorkWithFile(NameOfFile);
  WorkWithCNF(inFile);
   
//  free(MedianFlip);
//  free(MedianAssign);
  return -10;
}
#endif 
/////////////////////////////////////////////////////////////////////////////////////////////
#ifndef _SAT_COMP_ 
int main(int argc,char *argv[])
{
#ifdef _MPI_
  int myid, numprocs,namelen;
  char processor_name[MPI_MAX_PROCESSOR_NAME];
#endif
  number result, i;  
  char *NameOfFile;
  char *NameOfAssignFile = NULL;
  char *NameOfDir=".";
  int CheckInDir=1; 
  // will check all .cnf-files in a directory?
 
  struct timeval tv;
  struct timezone tz;

  char NameOfOutputFile[512];
#ifdef _MPI_
  MPI_Init(&argc,&argv);
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);
  MPI_Get_processor_name(processor_name,&namelen);

  fprintf(stderr,"c Process %lld on %s\n", myid, processor_name);
#endif
  if (argc==1) help();
  
  result = 28120107;
  srandom(result);
 
  //default values 
  I = DEF_I;
  T = DEF_T; 
  A = DEF_A;
  p = DEF_p;
  Quiet = 0;
  aGiven = 0;
  MaxTry = 1;
  *NameOfOutputFile = '\0'; 

  for (i=1;i<argc;i++)
  {
    if (strcmp(argv[i],"-o")==0)
    {
      *NameOfOutputFile = '\0';
      strcat(NameOfOutputFile,argv[++i]);
    }
    else if (strcmp(argv[i],"-sr")==0)
    {
      gettimeofday(&tv,&tz);
      result = tv.tv_usec;
      srandom(result);
    }
    else if (strcmp(argv[i],"-s")==0)
    {
      result = atoi(argv[++i]);
      srandom(result);
    }
    else if (strcmp(argv[i],"-f")==0)
    {
      NameOfFile = argv[++i];
      CheckInDir = 0; 
    }
    else if (strcmp(argv[i],"-d")==0)
      NameOfDir = argv[++i];
    else if (strcmp(argv[i],"-a")==0)
    {
      NameOfAssignFile = argv[++i];
      aGiven=1;
    }
    else if(strcmp(argv[i],"-I")==0)
         I = atoi(argv[++i]);
    else if(strcmp(argv[i],"-r")==0)
         MaxTry = atoi(argv[++i]);
    else if(strcmp(argv[i],"-T")==0)
        T = atoi(argv[++i]);
    else if(strcmp(argv[i],"-A")==0)
        A = atoi(argv[++i]);
    else if(strcmp(argv[i],"-p")==0)
        p = atoi(argv[++i]);
    else if(strcmp(argv[i],"-q")==0)
	Quiet = 1;
    else 
        help();
  }
  
  if (*NameOfOutputFile=='\0')
  {
    if (mkdir("Outputs",0xFFFF)) if (errno != EEXIST)
    {
      printf("Failed to create Outputs directory, and no -o key was given.\n");
      exit(-1);
    }
    strcat(NameOfOutputFile,"Outputs/");
    strcat(NameOfOutputFile,VERSION);
    if (mkdir(NameOfOutputFile,0xFFFF)) if (errno != EEXIST)
    {
      printf("Failed to create Outputs/VERSION directory, and no -o key was given.\n");
      exit(-1);
    }
    strcat(NameOfOutputFile,"/");
    strcat(NameOfOutputFile,VERSION);
    strcat(NameOfOutputFile,".");
    if (aGiven) 
    {
      strcat(NameOfOutputFile,basename(NameOfAssignFile)); 
      strcat(NameOfOutputFile,".a");
    }
    else if (CheckInDir) 
    { 
      strcat(NameOfOutputFile,basename(NameOfDir)); 
      strcat(NameOfOutputFile,".d"); 
    }
    else 
    {
      strcat(NameOfOutputFile,basename(NameOfFile)); 
      strcat(NameOfOutputFile,".f");
    }

    if (MaxTry>1)
    {
      strcat(NameOfOutputFile,".r");
      strcat(NameOfOutputFile,itoa(MaxTry));
    }
    if (p!=DEF_p)
    {
      strcat(NameOfOutputFile,".p");
      strcat(NameOfOutputFile,itoa(p));
    }
    if (T<DEF_T)
    {
      strcat(NameOfOutputFile,".T");
      strcat(NameOfOutputFile,itoa(T));
    }
    if (I<DEF_I)
    {
      strcat(NameOfOutputFile,".I");
      strcat(NameOfOutputFile,itoa(T));
    }
    if (A<DEF_A)
    {
      strcat(NameOfOutputFile,".A");
      strcat(NameOfOutputFile,itoa(A));
    }
    strcat(NameOfOutputFile,".s");
    strcat(NameOfOutputFile,itoa(result));
  
  
  if (Quiet) 
    strcat(NameOfOutputFile,".q");
  
  strcat(NameOfOutputFile,".pid");
  strcat(NameOfOutputFile,itoa(getpid()));
  
  }
  
  outFile  = fopen(NameOfOutputFile,"w");
  MedianFlip = (number*)malloc(MaxTry*sizeof(number));
  MedianAssign = (number*)malloc(MaxTry*sizeof(number));

  if(NameOfAssignFile!=NULL)
  {
    while(strcmp(NameOfOutputFile,NameOfAssignFile)==0)
      strcat(NameOfOutputFile,"1");
    
    fclose(outFile);
    outFile  = fopen(NameOfOutputFile,"w");
 
    WorkWithSatAssignment(NameOfAssignFile);
  }
  else if(CheckInDir)
    listdir(NameOfDir);
  else
  { 
    if (!Quiet) printf("f %s",NameOfFile);   
    fflush(stdout);
    fprintf(outFile,"f %s",NameOfFile);  
    fflush(outFile); 
    WorkWithFile(NameOfFile);
    WorkWithCNF(inFile);  
  }    
  
  if (!Quiet) printf("end\n");
  fprintf(outFile,"seed = %lld\n", result);
  fprintf(outFile,"end\n",NameOfFile);   
  fclose(outFile);
#ifdef _MPI_
  MPI_Finalize();
#endif
  return FAILS;
}
#endif
